/*
 * Transformations on STLC with recursion
 * Date Created:   October 15, 2014
*/

sig trans.

kind nat type.
type z nat.
type s nat -> nat.

type add   nat -> nat -> nat -> o.
type npred  nat -> nat -> o.
type is_nat  nat -> o.

% Source and target types are in the same category.
kind ty type.

type tnat    ty.                 % for natural numbers
type arr     ty -> ty -> ty.     % for abstractions in source and
                                 %% closures in the target languages.
type tunit   ty.                 % for units
type prod    ty -> ty -> ty.     % for pairs 
type arr'    ty -> ty -> ty.     % for abstractions in the target language


% Source Terms
kind tm type.

type nat         nat -> tm.
type pred        tm -> tm.
type plus        tm -> tm -> tm.
type ifz         tm -> tm -> tm -> tm.
type unit        tm.
type pair        tm -> tm -> tm.
type fst, snd    tm -> tm.
type let         tm -> (tm -> tm) -> tm.
type fix         (tm -> tm -> tm) -> tm.
type app         tm -> tm -> tm.

type bnd         (tm -> tm) -> tm.
type base        tm -> tm.

% Target Terms
kind tm' type.
    
type nat'         nat -> tm'.
type pred'        tm' -> tm'.
type plus'        tm' -> tm' -> tm'.
type ifz'         tm' -> tm' -> tm' -> tm'.
type unit'        tm'.
type pair'        tm' -> tm' -> tm'.
type fst', snd'   tm' -> tm'.
type let'         tm' -> (tm' -> tm') -> tm'.
type abs'         (tm' -> tm') -> tm'.
type app'         tm' -> tm' -> tm'.
type clos'         tm' -> tm' -> tm'.
type open'        tm' -> (tm' -> tm' -> tm') -> tm'.

% Terms
type tm          tm -> o.
type tm'         tm' -> o.
type tm''        tm' -> o.
type tm''_body   tm' -> o.

% Source term list
kind tm_list type.
type snil tm_list.
type scons tm -> tm_list -> tm_list.
type smember tm -> tm_list -> o.
type sappend tm_list -> tm_list -> tm_list -> o.

% Target term list
kind tm'_list type.
type cnil tm'_list.
type ccons tm' -> tm'_list -> tm'_list.
type cmember tm' -> tm'_list -> o.
type cappend tm'_list -> tm'_list -> tm'_list -> o.
type crev tm'_list -> tm'_list -> tm'_list -> o.

% Values
type val         tm -> o.
type val'        tm' -> o.
type tm'_list_val   tm'_list -> o.

% Typing rules
type of          tm -> ty -> o.
type ofc'        tm' -> ty -> o.
type of'         tm' -> ty -> o.
type of''        tm' -> ty -> o.
type of''_env    tm'_list -> ty -> o.
type of''_body   ty -> tm' -> ty -> o.

% Well-formed types
type is_sty       ty -> o.
type is_cty       ty -> o.

% Call-by-value evaluation rules in small steps
type step        tm -> tm -> o.
type nstep       nat -> tm -> tm -> o.
type eval        tm -> tm -> o.
type step'        tm' -> tm' -> o.
type nstep'       nat -> tm' -> tm' -> o.
type eval'        tm' -> tm' -> o.
type eval''       tm' -> tm' -> o.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%          CPS Transformation                 %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
type cps  tm -> (tm -> tm) -> tm -> o.
type cps' tm -> tm -> o.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%          Closure Conversion                 %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Mapping from source variables to target terms
kind map type.
type map tm -> tm' -> map.
kind map_list type.
type mnil map_list.
type mcons map -> map_list -> map_list.
type mmember map -> map_list -> o.
type mappend map_list -> map_list -> map_list -> o.

% Projection from free variables to an environment parameter
type mapvar tm_list -> (tm' -> map_list) -> o.

% Forming the closure environment
type mapenv tm_list -> map_list -> tm' -> o.

% Collecting free variables in an abstraction
type notfree tm -> o.
type fvars tm -> tm_list -> tm_list -> o.

% Combining two lists; avoiding redundancy is not essential for correctness
type combine tm_list -> tm_list -> tm_list -> o.

% Closure conversion
% The first argument is a mapping from free variables to target terms 
% and the second argument is a list of these free variables
type cc map_list -> tm_list -> tm -> tm' -> o.
type cc' tm -> tm' -> o.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%          Code Hoisting                      %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Hoisted terms consisting of a list of closed functions and the body
type hbase     tm' -> tm'.             % Base term of the body
type habs      (tm' -> tm') -> tm'.    % Top-level bindings of the body
type htm       tm'_list -> tm' -> tm'. % The hoisted term

% Combining the bodies of two hoisted terms
type hcombine      tm' -> tm' -> (tm' -> tm' -> tm') -> tm' -> o.
type hcombine3     tm' -> tm' -> tm' ->
                   (tm' -> tm' -> tm' -> tm') -> tm' -> o.
type hcombine_abs  tm' -> (tm' -> tm') -> 
                   (tm' -> (tm' -> tm') -> tm') -> tm' -> o.

% Apply the construct under the abstractions of 
% the body hoisted terms.
type hconstr       tm' -> (tm' -> tm') -> tm' -> o.

% Extract functions out of an abstraction and produce
% a closed abstraction
type hoist_abs     (tm' -> tm') -> tm' -> tm' -> tm'_list -> 
                   tm' -> (tm' -> tm') -> o.
type abstract   (tm' -> tm') -> (tm' -> tm') -> (tm' -> tm' -> tm') -> o.

type tm'_list_to_tuple     tm'_list -> tm' -> o.

% Code hoisting
type ch        tm' -> tm' -> o.